"""Snowflake Operators.

There is an existing snowflake_operator.py included with Airflow code. We are extending that module to support
features Snowflake does not include at the moment.
"""
from typing import Optional
from contextlib import closing

from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class MultiStatementSnowflakeOperator(BaseOperator):
    """Executes multiple sql statements.

    This Operator uses the `conn.execute_string` method of Snowflake Connector that allows running multiple sql
    statements.
    """

    template_fields = ('sql',)
    template_ext = ('.sql',)
    ui_color = '#ededed'

    @apply_defaults
    def __init__(self, sql: str, snowflake_conn_id: str = 'snowflake_default', parameters: Optional[dict] = None,
                 warehouse: Optional[str] = None, database: Optional[str] = None, *args, **kwargs):
        """Class Initializer.

        :param sql: the str of sql code to be executed. (templated)
        :type sql: A str representing a sql statement, or reference to a template file.
            Template reference are recognized by str ending in '.sql'
        :param parameters: extra information to process the queries
        :type parameters: dict
        :param snowflake_conn_id: reference to specific snowflake connection id
        :type snowflake_conn_id: str
        :param warehouse: name of warehouse which overwrite defined one in connection
        :type warehouse: str
        :param database: name of database which overwrite defined one in connection
        :type database: str
        """
        super().__init__(*args, **kwargs)
        self.snowflake_conn_id = snowflake_conn_id
        self.sql = sql
        self.parameters = parameters
        self.warehouse = warehouse
        self.database = database

    def get_hook(self) -> SnowflakeHook:
        """Get Hook."""
        return SnowflakeHook(
            snowflake_conn_id=self.snowflake_conn_id,
            warehouse=self.warehouse,
            database=self.database
        )

    def execute(self, context: dict) -> None:
        """Execute Multi Statement SQLs inside a Transaction."""
        self.log.info('Executing: %s', self.sql)

        with closing(self.get_hook().get_conn()) as conn:
            conn.execute_string(self.sql, params=self.parameters)



import os
from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="rappiflow_providers",
    version="0.3.5",
    author="CPGS DS Team",
    description='Rappiflow Dags and Operators utils',
    install_requires=requirements,
    packages=find_packages(),
    python_requires=">=3.7",
)



import os

from airflow.models import DAG
from airflow.operators.python import PythonOperator

from rappiflow_providers.operators.utils.time_execution import get_run_time_execution_table
from rappiflow_providers.operators.snowflake_operators import MultiStatementSnowflakeOperator

DAG_ID = "dag_metric_resources"
queries_base_path = os.path.join(os.path.dirname(__file__), 'queries')


def create_dag(
        schedule_interval,
        start_date,
        postgres_conn_id,
        snowflake_conn_id,
        interval,
        catchup,
        dev,
        rappiflow_instance_name
):
    schema = 'cpgs_datascience_dev' if dev else 'cpgs_datascience'
    params = {"rappiflow_name": rappiflow_instance_name,
              "times_execution_table": f"{schema}.rappiflow_dag_executions",
              "prometheus_table": f"{schema}.prometheus_pod_resources",
              "dm_dag_executions": f"{schema}.dm_rappiflow_dag_executions"}

    with DAG(
            dag_id=DAG_ID,
            schedule_interval=schedule_interval,
            max_active_runs=1,
            start_date=start_date,
            catchup=catchup,
            template_searchpath=queries_base_path,
            description="Get DAG metrics resource",
    ) as dag:
        get_dag_times_execution = PythonOperator(
            task_id="get_dag_times_execution",
            python_callable=get_run_time_execution_table,
            provide_context=True,
            op_kwargs={
                "table_name": params['times_execution_table'],
                "rappiflow_instance_name": rappiflow_instance_name,
                "postgres_conn_id": postgres_conn_id,
                "snowflake_conn_id": snowflake_conn_id,
                "interval": interval,
            },
        )

        dm_dag_executions_sql = 'load_dag_data.sql '
        load_daily_dag_executions_information = MultiStatementSnowflakeOperator(
            task_id='load_dag_executions_data',
            sql=dm_dag_executions_sql,
            snowflake_conn_id=snowflake_conn_id,
            params=params
        )

        get_dag_times_execution >> load_daily_dag_executions_information

        return dag



"""Time executions."""
from contextlib import closing
import logging

import pandas as pd
from airflow.providers.postgres.hooks.postgres import PostgresHook
from airflow.providers.snowflake.hooks.snowflake import SnowflakeHook
from snowflake.connector.pandas_tools import write_pandas

WAREHOUSE = "CPGS"
DATABASE = "FIVETRAN"


def get_run_time_execution_table(
        ds,
        postgres_conn_id,
        snowflake_conn_id,
        table_name,
        rappiflow_instance_name,
        **kwargs,
):
    """
    Connect to airflow database, gets DAG run executions data.

    1. Connect to Airflow Database
    2. Execute SELECT SQL statement on dag_run table
    3. Creates a Pandas DF from results on step 2
    4. Do some transformations
    5. Bulk data to Snowflake rappiflow_dag_executions table

    @return: None
    """
    # Read from Airflow database the DAG run executions times.
    columns = [
        "dag_id",
        "task_id",
        "owners",
        "execution_date",
        "start_date",
        "end_date",
        "state",
    ]

    sql_query = f"""
                  SELECT
                      ti.dag_id,
                      ti.task_id,
                      dag.owners,
                      ti.execution_date,
                      ti.start_date,
                      ti.end_date,
                      ti.state
                  FROM task_instance ti
                  INNER JOIN dag ON ti.dag_id = dag.dag_id
                  WHERE ti.execution_date::date = '{ds}'::date
                  """

    logging.info("Hooking Postgres Connection")
    postgres_hook = PostgresHook(postgres_conn_id=postgres_conn_id)

    logging.info("Getting records from airflow db: %s" % sql_query)
    records = postgres_hook.get_records(sql=sql_query)

    df = pd.DataFrame(records, columns=columns)
    df["rappiflow_name"] = rappiflow_instance_name
    df["run_time_hours"] = df["end_date"] - df["start_date"]
    df["run_time_hours"] = df["run_time_hours"].apply(
        lambda x: "{:.4f}".format(x / pd.Timedelta("1 hour")) if pd.notnull(x) else x
    )

    logging.info("Dataframe Shape: %s" % str(df.shape))

    # When there's not data on end_date, pandas adds a NaT value that brakes the insertion.
    # To avoid it just replace those values with None.
    df = df.astype(str).replace({"NaT": None})

    logging.info("Hooking Snowflake Connection")
    schema = table_name.split('.')[0]
    snowflake_hook = SnowflakeHook(
        snowflake_conn_id=snowflake_conn_id,
        warehouse=WAREHOUSE,
        database=DATABASE,
        schema=schema
    )

    logging.info("Bulking data into Snowflake")
    with closing(snowflake_hook.get_conn()) as conn:
        cursor = conn.cursor()
        delete_query = f"""
                          DELETE FROM {table_name}
                          WHERE to_date(start_date) = '{ds}'
                          AND RAPPIFLOW_NAME = '{rappiflow_instance_name}'
                          """

        logging.info("Deleting Data: %s" % delete_query)
        cursor.execute(delete_query)
        write_pandas(conn=conn, df=df, table_name=table_name, quote_identifiers=False)
